#include "bilinear.h"
#include "ImageMorpher.h"
#include "LineSegment.h"
#include "PixelMap.h"

#include <cmath>
#include <svl/SVL.h>
#include <cassert>
#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
#include <pthread.h>
#include <string>
using std::string;

namespace
{
    /**
     * Structure passed to MorphRows as a thread argument.
     */
    typedef struct
    {
        int startRow, endRow;
        CImageMorpher *pThis;
    } MorphRowsArg;

    /**
     * Structure passed to BGMorph as a thread argument.
     */
    typedef struct
    {
        MorphInfo info;
        CImageMorpher *pThis;
    } BGMorphArg;
}

static void DrawLine(CPixelMap* img, int x0, int y0, int x1, int y1, Pixel8 pix[3])
{
    int dy = y1 - y0;
    int dx = x1 - x0;
    float t = (float) 0.5f;

    img->SetPixel4(x0, y0, pix);
    if (abs(dx) > abs(dy))
    {
        float m = (float) dy / (float) dx;
        t += y0;
        dx = (dx < 0) ? -1 : 1;
        m *= dx;
        while (x0 != x1)
        {
            x0 += dx;
            t += m;
            img->SetPixel4(x0, (int)t, pix);
        }
    }
    else
    {
        float m = (float) dx / (float) dy;      // compute slope
        t += x0;
        dy = (dy < 0) ? -1 : 1;
        m *= dy;
        while (y0 != y1)
        {
            y0 += dy;
            t += m;
            img->SetPixel4((int)t, y0, pix);
        }
    }

    Pixel8 endColor[3] = {0, 255, 255};
    img->SetPixel4(x1, y1, endColor);
}


CImageMorpher::CImageMorpher(void)
{
    Reset();
    m_numThreads = 1;
    m_completionCallback = NULL;
    m_completionCallbackArg = NULL;
    pthread_mutex_init(&m_bgMorphMutex, NULL);
    pthread_mutex_init(&m_morphMutex, NULL);
    pthread_mutex_init(&m_morphStatusMutex, NULL);
}


CImageMorpher::~CImageMorpher(void)
{
    // If a morph is still running in the background,
    // kill it!
    float status;
    GetMorphStatus(&status);
    if (status < 1.0f)
      pthread_cancel(m_bgMorphThread);

    Reset();
    pthread_mutex_destroy(&m_bgMorphMutex);
    pthread_mutex_destroy(&m_morphMutex);
    pthread_mutex_destroy(&m_morphStatusMutex);
}


void CImageMorpher::SetNumThreads(const unsigned int numThreads)
{
    if (numThreads > 0)
        m_numThreads = numThreads;
}


int CImageMorpher::Init(const MorphInfo &info)
{
    Reset();

    // Process MorphInfo
    m_warpFrac = info.warpFrac;
    m_dissolveFrac = info.dissolveFrac;
    m_morphParamA = info.morphParamA;
    m_morphParamB = info.morphParamB;
    m_morphParamP = info.morphParamP;
    m_outWidth  = info.outWidth;
    m_outHeight = info.outHeight;
    m_imgA = info.imgA;
    m_imgB = info.imgB;
    m_imgOut = info.imgOut;

    m_originalLinesA = info.linesA;
    m_originalLinesB = info.linesB;
    PreprocessLines();

    // Test validity of images and fractions.
    if (m_warpFrac < 0.0f || m_warpFrac > 1.0f)
    {
        cerr << "ERROR: warpFrac (" << m_warpFrac << ") must be in the range [0..1]\n";
        return -1;
    }
    if (m_dissolveFrac < 0.0f || m_dissolveFrac > 1.0f)
    {
        cerr << "ERROR: dissoveFrac (" << m_dissolveFrac << ") must be in the range [0..1]\n";
        return -1;
    }
    if ((m_imgA->GetBytesPerPixel() != 4) || (m_imgB->GetBytesPerPixel() != 4))
    {
        cerr << "ERROR: " << m_imgA->GetSourceFile() << " and " << m_imgB->GetSourceFile()
            << " must be 32-bit images!\n";
        return -1;
    }
    return 0;
}


void CImageMorpher::PreprocessLines(void)
{
    int i;

    assert(m_originalLinesA != NULL && m_originalLinesB != NULL);
    const SegmentVector& originalsA = *m_originalLinesA;
    const SegmentVector& originalsB = *m_originalLinesB;
    assert(originalsA.size() == originalsB.size());
    int numLines = (int)originalsA.size();
    if (numLines == 0)
        return; // not an error, just a really boring morph.

    // clean up any existing arrays
    m_linesA.clear();
    m_linesA.reserve(numLines);
    m_linesB.clear();
    m_linesB.reserve(numLines);
    m_linesInterp.clear();
    m_linesInterp.reserve(numLines);

    // fill new arrays
    Vec2 pvecEndpoints[2];
    float points[4];
    int widthA = m_imgA->GetWidth();
    int heightA = m_imgA->GetHeight();
    int widthB = m_imgB->GetWidth();
    int heightB = m_imgB->GetHeight();
    for(i=0; i<numLines; ++i)
    {
        // assign scaled points to segments A.  Remember to flip
        // the Y coordinates.
        points[0] = originalsA[i].GetAX();
        points[1] = (1.0f - originalsA[i].GetAY());
        points[2] = originalsA[i].GetBX();
        points[3] = (1.0f - originalsA[i].GetBY());
        m_linesA.push_back(CLineSegment());
        m_linesA[i].SetPoints(points);
        m_linesA[i].Precalculate(m_morphParamP);

        // assign scaled points to segment B.  Remember to flip
        // the Y coordinates.
        points[0] = originalsB[i].GetAX();
        points[1] = (1.0f - originalsB[i].GetAY());
        points[2] = originalsB[i].GetBX();
        points[3] = (1.0f - originalsB[i].GetBY());
        m_linesB.push_back(CLineSegment());
        m_linesB[i].SetPoints(points);
        m_linesB[i].Precalculate(m_morphParamP);

        // Calculate interpolated value.
        CLineSegment::Interpolate(m_linesA[i], m_linesB[i], m_warpFrac,
            pvecEndpoints);
        m_linesInterp.push_back(CLineSegment());
        m_linesInterp[i].SetPoints(pvecEndpoints[0], pvecEndpoints[1]);
        m_linesInterp[i].Precalculate(m_morphParamP);
    }
}


void CImageMorpher::Reset(void)
{
    m_imgA = NULL;
    m_imgB = NULL;
    m_imgOut = NULL;
    m_originalLinesA = NULL;
    m_originalLinesB = NULL;
    m_numRowsFinished = 0;
}


int CImageMorpher::Morph(const MorphInfo &info)
{
    pthread_mutex_lock(&m_morphMutex);
    
    if (Init(info) != 0)
    {
        Reset();
        pthread_mutex_unlock(&m_morphMutex);
        return -1;
    }

    // create picOut, at long last
    m_imgOut->Initialize(m_outWidth, m_outHeight, 4);

    // TEMP: line debugging
    /**
    Pixel8 color[3] = {255, 255, 0};
    memset(m_imgOut->GetPixels(), 0, m_imgOut->GetWidth() * m_imgOut->GetHeight() * 4);
    const SegmentVector& lines = m_linesInterp;
    for(i=0; i<lines.size(); ++i)
    {
        DrawLine(m_imgOut, (int)lines[i].GetAX(), (int)lines[i].GetAY(),
            (int)lines[i].GetBX(), (int)lines[i].GetBY(), color);
    }
    pthread_mutex_unlock(&m_morphMutex);
    return 0;
    /**/

    // divide up the work among the worker threads.
    int numThreads = m_numThreads; // this could change in mid-morph if we're not careful!
    int rowsPerThread = (int)floor((float)m_outHeight / (float)numThreads);
    MorphRowsArg *threadArgs = new MorphRowsArg[numThreads];
    pthread_t *threads = new pthread_t[numThreads];
    int i;
    for(i=0; i<numThreads; ++i)
    {
        threadArgs[i].startRow = i*rowsPerThread;
        threadArgs[i].endRow   = (i == numThreads-1) ? m_outHeight : (i+1)*rowsPerThread;
        threadArgs[i].pThis    = this;   
        int err = pthread_create(&threads[i], NULL, MorphRows,
            (void *)(&threadArgs[i]));
        if (err != 0)
        {
            cerr << "Thread creation error in __FUNCTON__\n";
            // TODO: better error handling here.
        }
    }

    // join the threads when they're finished.
    for(i=0; i<numThreads; ++i)
    {
        pthread_join(threads[i], NULL);
    }
    delete [] threadArgs;
    delete [] threads;

    if (m_completionCallback != NULL)
    {
        m_completionCallback(m_completionCallbackArg);
    }

    Reset();
    pthread_mutex_unlock(&m_morphMutex);
    return 0;
}

int CImageMorpher::MorphInBackground(const MorphInfo &info)
{
    pthread_mutex_lock(&m_bgMorphMutex);
    BGMorphArg *threadArgs = new BGMorphArg;
    threadArgs->info = const_cast<MorphInfo&>(info);
    threadArgs->pThis = this;
    int err = pthread_create(&m_bgMorphThread, NULL, BGMorph, (void*)(threadArgs));
    if (err != 0)
    {
        cerr << "Thread creation error in __FUNCTION__\n";
        // TODO: better error handling here.
    }
    pthread_mutex_unlock(&m_bgMorphMutex);
    return 0;
}


int CImageMorpher::GetMorphStatus(float *outStatus)
{
    if (outStatus == NULL)
        return -1;
    if (m_outHeight <= 0)
        return -1;
    
    // See if there's a morph in progress at all
    if (pthread_mutex_trylock(&m_morphMutex) == EBUSY)
    {
        // there's a morph in progress.
        *outStatus = (float)m_numRowsFinished / (float)m_outHeight;
    }
    else
    {
        // no morph in progress.  Release the mutex, quickly-like!
        pthread_mutex_unlock(&m_morphMutex);
        *outStatus = 1.0f;
    }
    return 0;
}


// FINDXPRIME -- finds the X' in a source image that corresponds to X
// in the destination image.  Stores X' in the *Xp vector.  Returns 0
// if all is well, anything else if it's not.
void CImageMorpher::FindXPrimes(const Vec2 &X, Vec2 *XpA, Vec2 *XpB,
    const bool checkAllWeights, bool usedWeights[])
{
    Vec2 DsumA(0.0f, 0.0f), DsumB(0.0f, 0.0f);
    float dist, weight, weightsumA = 0.00001f, weightsumB = 0.00001f;
    unsigned int i;

    // loop over pairs of segments
    for(i=0; i<m_linesInterp.size(); ++i)
    {
        // if we're not checking all weights this pixel, and the array
        // indicates that this line's weight wasn't significant, don't
        // bother checking it.
        if (!checkAllWeights && !usedWeights[i])
            continue;

        // calculate u, v based on interpSegments[i]
        float u,v;
        m_linesInterp[i].GetUV(X, &u, &v);

        // calculate weight from this segment/point
        dist = m_linesInterp[i].DistanceToPoint(X, u, v);
        weight = (float)pow(m_linesInterp[i].GetLengthP() / (m_morphParamA+dist),
            m_morphParamB);

        // update Dsum and weightsum
        if (weight >= 0.001f)
        {
            Vec2 Xp_tempA, Xp_tempB, DA, DB;
            // calculate Xp_temp and the displacement D.  We must test to
            // see if Xp_temp is within the bounds of the image; if not, we
            // disregard this whole loop iteration.
            m_linesA[i].GetX(u, v, &Xp_tempA);
            if ( (Xp_tempA[0] < 0.0f) || (Xp_tempA[0] > 1.0f) ||
                 (Xp_tempA[1] < 0.0f) || (Xp_tempA[1] > 1.0f) )
                continue;
            m_linesB[i].GetX(u, v, &Xp_tempB);
            if ( (Xp_tempB[0] < 0.0f) || (Xp_tempB[0] > 1.0f) ||
                 (Xp_tempB[1] < 0.0f) || (Xp_tempB[1] > 1.0f) )
                continue;

            DA = Xp_tempA - X;
            DsumA += weight*DA;
            weightsumA += weight;
            DB = Xp_tempB - X;
            DsumB += weight*DB;
            weightsumB += weight;

            usedWeights[i] = true;
        }
        else
        {
            usedWeights[i] = false;
        }
    }
    *XpA = X + (DsumA/weightsumA);
    *XpB = X + (DsumB/weightsumB);
}

// MORPHROWS -- determines the final colors for pixels in the given
// rows. This is the work performed by one thread.
void *CImageMorpher::MorphRows(void *t_args)
{
    MorphRowsArg *realArgs = (MorphRowsArg *)t_args;
    int start = realArgs->startRow, end = realArgs->endRow;
    CImageMorpher *pThis = realArgs->pThis;
    int x, y;
    int nx = pThis->m_outWidth;

    // allocate some mega-time-saving arrays
    bool *usedWeights = new bool[pThis->m_linesA.size()];
    if (usedWeights == NULL)
    {
        cerr << "ERROR: memory allocation error in " << __FUNCTION__ << endl;
        return NULL;
    }

    unsigned char colorA[3];
    unsigned char colorB[3];
    int widthA  = pThis->m_imgA->GetWidth();
    int widthB  = pThis->m_imgB->GetWidth();
    int heightA = pThis->m_imgA->GetHeight();
    int heightB = pThis->m_imgB->GetHeight();
    float widthOut  = (float)pThis->m_outWidth;
    float heightOut = (float)pThis->m_outHeight;
    float stepX = 1.0f / widthOut;
    float stepY = 1.0f / heightOut;
    Vec2 X(0.5f/widthOut, (float)(start+0.5f)/heightOut), Xp_a, Xp_b;

    // advance to the start of the appropriate row of pixels
    Pixel8 *pixel = pThis->m_imgOut->GetPixels();
    pixel += 4*(start*nx);
    for(y=start; y<end; ++y)
    {
        X[0] = 0.5f/widthOut;
        for(x=0; x<nx; ++x)
        {
            // are we checking all weights?  TO DO: change things such
            // that you check one pixel and then check NEIGHBORING
            // pixels, instead of all in a row.  that might be difficult
            // though.
            //bool checkAllWeights = true; // recalc every pixel
            bool checkAllWeights = !(x & 0x3); // recalc every 4 pixels
            //bool checkAllWeights = !(x & 0x7); // recalc every 8 pixels

            // find coordinates of corresponding pixels in A and B
            pThis->FindXPrimes(X, &Xp_a, &Xp_b, checkAllWeights, usedWeights);
            
            // get colors of Xp_A and Xp_B using bilinear interpolation
            bilinear(*(pThis->m_imgA), (float)(Xp_a[0]*widthA), (float)(Xp_a[1]*heightA), colorA);
            bilinear(*(pThis->m_imgB), (float)(Xp_b[0]*widthB), (float)(Xp_b[1]*heightB), colorB);

            // calculate and store the final color for this pixel!
            pixel[0] = (unsigned char)( colorA[0] + pThis->m_dissolveFrac * (colorB[0]-colorA[0]));
            pixel[1] = (unsigned char)( colorA[1] + pThis->m_dissolveFrac * (colorB[1]-colorA[1]));
            pixel[2] = (unsigned char)( colorA[2] + pThis->m_dissolveFrac * (colorB[2]-colorA[2]));
            pixel[3] = 0xFF; // alpha
            pixel += 4;
            X[0] += stepX;
        }
        X[1] += stepY;

        // Update the number of completed rows
        pthread_mutex_lock(&pThis->m_morphStatusMutex);
        pThis->m_numRowsFinished++;
        pthread_mutex_unlock(&pThis->m_morphStatusMutex);

#ifndef NDEBUG  
        if (end == pThis->m_imgOut->GetHeight())
        {
            float status;
            pThis->GetMorphStatus(&status);
            printf("%.2f%% finished\n", status*100.0f);
        }
#endif
    }
    delete [] usedWeights;
    return NULL;
}

void *CImageMorpher::BGMorph(void *t_args)
{
    BGMorphArg *realArgs = (BGMorphArg *)t_args;
    MorphInfo info = realArgs->info;
    CImageMorpher *pThis = realArgs->pThis;
    delete realArgs;

    pThis->Morph(info);

    pthread_exit(NULL);
    return NULL;
}


void CImageMorpher::SetCompleteCallback(void (*callbackFunc)(void*), void *arg)
{
    m_completionCallback = callbackFunc;
    m_completionCallbackArg = arg;
}
