#include "ImageMorpher.h"
#include "LineEditor.h"
#include "MorphView.h"

#pragma warning(push)
#pragma warning(disable:4311 4312)
#include <FL/Fl.H>
#include <FL/fl_ask.H>
#pragma warning(pop)

#if WIN32
#include <windows.h>
#endif
#include <GL/gl.h>

#include <cassert>

namespace
{
    const float CLOSE_LIMIT = 0.04f;
    const float BIG_VALUE = 1000.0f;
    
    const float LINE_COLOR[3] = {0.0f, 1.0f, 0.0f};
    const float SELECTED_LINE_COLOR[3] = {1.0f, 0.0f, 0.0f};
}

const CLineEditor::LineID CLineEditor::INVALID_SEG = -1;


CLineEditor::CLineEditor(void)
{
    m_lineListA.clear();
    m_lineListB.clear();
    m_selectedSegmentId = INVALID_SEG;
    m_selectedSegmentPoint = 0;
    m_mouseButtonState[0] = m_mouseButtonState[1] =
        m_mouseButtonState[2] = BUTTON_STATE_NONE;
}

CLineEditor::~CLineEditor(void)
{
    m_lineListA.clear();
    m_lineListB.clear();
}


CLineEditor::LineID CLineEditor::AddLine(const Vec2& a, const Vec2& b)
{
    CLineSegment newLine;
    newLine.SetPoints(a, b);
    m_lineListA.push_back(newLine);
    m_lineListB.push_back(newLine);
    return (LineID)(m_lineListA.size()-1);
}


bool CLineEditor::DeleteLine(const LineID id)
{
    if (id == INVALID_SEG || id >= (int)m_lineListA.size())
        return false;
    SegmentVectorItor itor = m_lineListA.begin() + id;
    m_lineListA.erase(itor);
    itor = m_lineListB.begin() + id;
    m_lineListB.erase(itor);
    return true;
}


void CLineEditor::UpdateLine(const CMorphView::ImageID whichImg,
                             const LineID whichLine, Vec2 *newStart,
                             Vec2 *newEnd)
{
    assert(whichImg == CMorphView::IMG_A || whichImg == CMorphView::IMG_B);
    assert(whichLine < (int)m_lineListA.size());
    if (whichLine == INVALID_SEG)
        return;
    
    SegmentVector& lineList = (whichImg == CMorphView::IMG_A) ? m_lineListA : m_lineListB;
    if (newStart != NULL)
        lineList[whichLine].SetA(*newStart);
    if (newEnd != NULL)
        lineList[whichLine].SetB(*newEnd);
}


bool CLineEditor::GetClosestPoint(const CMorphView::ImageID whichImg,
                                  const float x, const float y,
                                  LineID *whichLine, int *whichPoint) const
{ 
    assert(whichImg == CMorphView::IMG_A || whichImg == CMorphView::IMG_B);
    unsigned int i, tempPt = 0;
    float tempDist; 
    LineID tempSeg = INVALID_SEG; 
    float dist = BIG_VALUE; 

    const SegmentVector& lineList = (whichImg == CMorphView::IMG_A)
        ? m_lineListA : m_lineListB;
    for(i=0; i<lineList.size(); ++i)
    { 
        const CLineSegment &line = lineList[i];

        float ax = line.GetAX();
        float ay = line.GetAY();
        tempDist = (ax-x)*(ax-x) + (ay-y)*(ay-y);
        if (tempDist < dist)
        {
            dist = tempDist;
            tempSeg = i;
            tempPt = 0;
        }

        float bx = line.GetBX();
        float by = line.GetBY();
        tempDist = (bx-x)*(bx-x) + (by-y)*(by-y);
        if (tempDist < dist)
        {
            dist = tempDist;
            tempSeg = i;
            tempPt = 1;
        }
    } 

    if(dist<CLOSE_LIMIT)
    {
        *whichLine = tempSeg;
        *whichPoint = tempPt;
        return true;
    }
    else
        return false; 
}


void CLineEditor::DrawLines(const CMorphView::ImageID whichImg,
                            const Vec2& min, const Vec2& max) const
{
    if (whichImg == CMorphView::IMG_OUT)
        return;

    // draw control lines
    const SegmentVector& lineList = (whichImg == CMorphView::IMG_A)
        ? m_lineListA : m_lineListB;
    unsigned int i;
    float width = max[0] - min[0] + 1.0f;
    float height = max[1] - min[1] + 1.0f;
    glColor3fv(LINE_COLOR);
    glPushMatrix();
    glTranslatef(min[0], max[1], 0.0f);
    glScalef(1.0f, -1.0f, 1.0f);
    for(i=0; i<lineList.size(); ++i)
    {
        const CLineSegment& line = lineList[i];
        float ax,ay; 
        float bx,by;

        ax = (line.GetAX() * width)  - 0.5f;
        ay = (line.GetAY() * height) + 0.5f;
        bx = (line.GetBX() * width)  - 0.5f;
        by = (line.GetBY() * height) + 0.5f;

        // draw the selected line in a different color
        if (i == m_selectedSegmentId)
            glColor3fv(SELECTED_LINE_COLOR);

        glBegin(GL_LINES); 
        glVertex2f(ax, ay); 
        glVertex2f(bx, by); 
        glEnd(); 

        glPointSize(4); 

        glBegin(GL_POINTS); 
        glVertex2f(bx, by); 
        glEnd();

        // reset color to default.
        if (i == m_selectedSegmentId)
            glColor3fv(LINE_COLOR);
    }

    glPopMatrix();
}


void CLineEditor::GatherMorphInfo(MorphInfo *info)
{
    info->linesA = &m_lineListA;
    info->linesB = &m_lineListB;
}


void CLineEditor::DisperseMorphInfo(const MorphInfo *info)
{
    m_lineListA.clear();
    m_lineListB.clear();
    m_lineListA.insert(m_lineListA.begin(), info->linesA->begin(), info->linesA->end());
    m_lineListB.insert(m_lineListB.begin(), info->linesB->begin(), info->linesB->end());
    m_selectedSegmentId = INVALID_SEG;
    m_selectedSegmentPoint = 0;
}


int CLineEditor::HandleMousePush(const CMorphView::ImageID whichImg,
                                 const CMorphView::ToolID whichTool,
                                 const float mx, const float my)
{
    if (whichImg == CMorphView::IMG_OUT)
        return 0;
    if ((Fl::event_state() & FL_BUTTON1) == 0)
        return 0;

    m_mouseButtonState[0] = BUTTON_STATE_PRESSED;
    if (whichTool == CMorphView::TOOL_ADD)
    {
        // Add a new line segment where the user clicked.
        Vec2 pt(mx,my);
        m_selectedSegmentId = AddLine(pt, pt);
        if (m_selectedSegmentId == INVALID_SEG)
        {
            m_mouseButtonState[0] = BUTTON_STATE_NONE;
            fl_alert("Error adding line segment.");
        }
        return 1;
    }
    else if (whichTool == CMorphView::TOOL_DELETE)
    {
        // Delete the closest line.
        m_selectedSegmentId = INVALID_SEG;   
        bool found = GetClosestPoint(whichImg, mx, my,
            &m_selectedSegmentId, &m_selectedSegmentPoint);
        if (found)
        {
            DeleteLine(m_selectedSegmentId);
            m_selectedSegmentId = INVALID_SEG;
        }
        return 1;
    }
    else if (whichTool == CMorphView::TOOL_EDIT)
    {
        // Find the closest line segment to the mouse,
        // and mark it as the current line for editing.
        bool found = GetClosestPoint(whichImg, mx, my, &m_selectedSegmentId,
            &m_selectedSegmentPoint);
        if (!found)
        { 
            m_mouseButtonState[0] = BUTTON_STATE_NONE;
            m_selectedSegmentId = INVALID_SEG;
            m_selectedSegmentPoint = 0; 
            fl_alert("No closer line segments,please click closer to the line segment."); 
        }
        return 1;
    }
    return 0;
}


int CLineEditor::HandleMouseDrag(const CMorphView::ImageID whichImg,
                                 const CMorphView::ToolID whichTool,
                                 const float mx, const float my) 
{
    if (whichImg == CMorphView::IMG_OUT)
        return 0;
    if ((Fl::event_state() & FL_BUTTON1) == 0)
        return 0;

    Vec2 newPt(mx,my);
    if (whichTool == CMorphView::TOOL_ADD)
    { 
        // Dragging the endpoint of a new line.
        if (m_mouseButtonState[0] != BUTTON_STATE_NONE)
        { 
            m_mouseButtonState[0] = BUTTON_STATE_HELD;

            // Update position of dragged point
            UpdateLine(CMorphView::IMG_A, m_selectedSegmentId, NULL, &newPt);
            UpdateLine(CMorphView::IMG_B, m_selectedSegmentId, NULL, &newPt);
        }
        return 1;
    }
    else if (whichTool == CMorphView::TOOL_EDIT)
    { 
        // Dragging selected point of an existing line.
        if (m_mouseButtonState[0] != BUTTON_STATE_NONE)
        {
            m_mouseButtonState[0] = BUTTON_STATE_HELD;
            if (m_selectedSegmentPoint == 0)
                UpdateLine(whichImg, m_selectedSegmentId, &newPt, NULL);
            else if (m_selectedSegmentPoint == 1)
                UpdateLine(whichImg, m_selectedSegmentId, NULL, &newPt);

            // If the shift key is held, update the position of the point in
            // the other image as well.
            if(Fl::event_state()&FL_SHIFT)
            { 
                CMorphView::ImageID otherImage = (CMorphView::ImageID)(1 - whichImg);
                if (m_selectedSegmentPoint == 0)
                    UpdateLine(otherImage, m_selectedSegmentId, &newPt, NULL);
                else if (m_selectedSegmentPoint == 1)
                    UpdateLine(otherImage, m_selectedSegmentId, NULL, &newPt);
            } 
        }
        else
        { 
            fl_alert("Invalid drag, not start from inside of a image.\n"); 
        }
        return 1;
    }
    else if (whichTool == CMorphView::TOOL_DELETE)
    { 
        // Dragging when deleting has no effect.
        return 1;
    }
    return 0;
}


int CLineEditor::HandleMouseRelease(const CMorphView::ImageID whichImg,
                                    const CMorphView::ToolID whichTool,
                                    const float mx, const float my) 
{
    if (whichImg == CMorphView::IMG_OUT)
        return 0;
    if ((Fl::event_state() & FL_BUTTON1) == 0)
        return 0;

    Vec2 newPt(mx, my);
    if (whichTool == CMorphView::TOOL_ADD)
    { 
        if(m_mouseButtonState[0] == BUTTON_STATE_HELD)
        { 
            // Stop dragging endpoint of new line.
            // Update the coordinates of the point the user released in both images.
            UpdateLine(CMorphView::IMG_A, m_selectedSegmentId, NULL, &newPt);
            UpdateLine(CMorphView::IMG_B, m_selectedSegmentId, NULL, &newPt);
        }

        // Update UI state.
        //m_currentImageId = CMorphView::IMG_OUT;
        m_selectedSegmentId = INVALID_SEG; 
        m_mouseButtonState[0] = BUTTON_STATE_NONE;
        return 1;
    }
    else if (whichTool == CMorphView::TOOL_EDIT)
    { 
        if(m_mouseButtonState[0] == BUTTON_STATE_HELD)
        { 
            // end dragging selected point of existing line.
            // Update the appropriate point in the current image.
            if (m_selectedSegmentPoint == 0)
                UpdateLine(whichImg, m_selectedSegmentId, &newPt, NULL);
            else if (m_selectedSegmentPoint == 1)
                UpdateLine(whichImg, m_selectedSegmentId, NULL, &newPt);

            // If the shift key is held, update the position of the point in
            // the other image as well.
            if(Fl::event_state()&FL_SHIFT)
            { 
                CMorphView::ImageID otherImage = (CMorphView::ImageID)(1 - whichImg);
                if (m_selectedSegmentPoint == 0)
                    UpdateLine(otherImage, m_selectedSegmentId, &newPt, NULL);
                else if (m_selectedSegmentPoint == 1)
                    UpdateLine(otherImage, m_selectedSegmentId, NULL, &newPt);
            }
        }
        else
        { 
            // end click while editing.
            // do nothing.
        } 

        // Update UI state.
        // m_currentImageId = CMorphView::IMG_OUT;
        m_selectedSegmentId = INVALID_SEG;
        m_mouseButtonState[0] = BUTTON_STATE_NONE;
        return 1;
    }
    else if (whichTool == CMorphView::TOOL_DELETE)
    {
        // mouse release events have no effect when deleting lines.
        return 1;
    }
    return 0;
}
