#ifndef LINEEDITOR_H
#define LINEEDITOR_H

#include "LineSegment.h"
#include "MorphView.h"

struct MorphInfo;

#include <svl/SVL.h>

class CLineEditor
{
public:
    CLineEditor(void);
    ~CLineEditor(void);

    /**
     * Draw lines.
     * @param whichImg Which image to draw the lines on top of.
     * @param min Coordinates of the minimum point of the image (closest to the
     *            origin).
     * @param max Coordinates of the maximum point of the image (farthest from
     *            the origin.
     */
    void DrawLines(const CMorphView::ImageID whichImg, const Vec2& min,
        const Vec2& max) const;

    void GatherMorphInfo(MorphInfo *info);
    void DisperseMorphInfo(const MorphInfo *info);

    int HandleMousePush(const CMorphView::ImageID whichImg,
        const CMorphView::ToolID whichTool, const float mx, const float my); 
    int HandleMouseDrag(const CMorphView::ImageID whichImg,
        const CMorphView::ToolID whichTool, const float mx, const float my); 
    int HandleMouseRelease(const CMorphView::ImageID whichImg,
        const CMorphView::ToolID whichTool, const float mx, const float my);

private:
    typedef int LineID;
    static const LineID INVALID_SEG;

    /**
     * Add a new line at the specified coordinates (in both images).
     * @param a Point A.
     * @param b Point B.
     * @return The id of the new line.
     */
    LineID AddLine(const Vec2& a, const Vec2& b);

    /**
     * Removes the line with the specified ID.  This has the side effect of
     * invalidating all existing LineIDs.
     * @param id The ID of the line to delete.
     * @return True if the line was deleted, false otherwise.
     */
    bool DeleteLine(const LineID id);
    
    /**
     * Find the point/line closest to the given coordinates (usually the
     * location of a mouse click.
     * @param whichImg Which image to look in.
     * @param x The x coordinate to search from.
     * @param y The y coordinate to search from.
     * @param whichLine If found, the ID of the nearest line will be put here.
     * @param whichPoint If a line is found, the value pointed to by whichPoint
     *                   will be 0 if the start/A point was closer, and 1 if
     *                   the end/B point was closer.
     * @return true if a line was found, false otherwise.
     */
    bool GetClosestPoint(const CMorphView::ImageID whichImg, const float x,
        const float y, LineID *whichLine, int *whichPoint) const;

    /**
     * Update the endpoints of a line.
     * @param whichImg Which image is the line in?
     * @param whichLine Which line to update?
     * @param newStart Coordinates for new starting point.  If NULL, the
                       starting point is unchanged.
     * @param newEnd Coordinates for new ending point.  If NULL, the
                     ending point is unchanged.
     */
    void UpdateLine(const CMorphView::ImageID whichImg, const LineID whichLine,
        Vec2 *newStart, Vec2 *newEnd);

    SegmentVector m_lineListA, m_lineListB;

    LineID m_selectedSegmentId;
    int m_selectedSegmentPoint; /**< 0 means point A, 1 means point B */

    typedef enum
    {
        BUTTON_STATE_NONE = 0,
        BUTTON_STATE_PRESSED,
        BUTTON_STATE_HELD,
    } ButtonState;
    ButtonState m_mouseButtonState[3]; /**< 0 = left, 1 = middle, 2 = right */
};

#endif // LINEEDITOR_H
