#include "LineSegment.h"
#include "fastmath.h"

#include <cassert>
#include <svl/SVL.h>

/////////////////////////////////////////////////////////////////////
// Construction/Destruction
/////////////////////////////////////////////////////////////////////

CLineSegment::CLineSegment(void)
{
    // No default initialization.
}
CLineSegment::CLineSegment(const CLineSegment &rhs)
{
    memcpy(this, &rhs, sizeof(CLineSegment));
}

CLineSegment::~CLineSegment()
{

}



/////////////////////////////////////////////////////////////////////
// Overloaded Operators
/////////////////////////////////////////////////////////////////////

// operator==
bool CLineSegment::operator==(const CLineSegment& rhs)
{
    return ( (m_vecA == rhs.m_vecA) && (m_vecB ==rhs.m_vecB) );
}



// operator=
CLineSegment& CLineSegment::operator=(const CLineSegment& rhs)
{
    // Test for self-assignment.
    if (*this == rhs)
        return *this;

    memcpy(this, &rhs, sizeof(CLineSegment));
    return *this;
}


/////////////////////////////////////////////////////////////////////
// Public Methods
/////////////////////////////////////////////////////////////////////

// SetPoints
void CLineSegment::SetPoints(const float fAX, const float fAY,
                             const float fBX, const float fBY)
{
    this->SetPoints( Vec2(fAX, fAY), Vec2(fBX, fBY) );
}
void CLineSegment::SetPoints(const Vec2& vecA, const Vec2& vecB)
{
    m_vecA = vecA;
    m_vecB = vecB;
}
void CLineSegment::SetPoints(const float pfPoints[4])
{
    this->SetPoints( Vec2(pfPoints[0], pfPoints[1]),
        Vec2(pfPoints[2], pfPoints[3]) );
}


// GetUV
void CLineSegment::GetUV(const Vec2 &vecX, float *pfU, float *pfV) const
{
    // calculate the U and V coordinates
    Vec2 vecXminusA(vecX - m_vecA);
    *pfU = static_cast<float>(dot(vecXminusA, m_vecAB) * m_fInverseLength2);
    *pfV = static_cast<float>(dot(vecXminusA, m_vecPerpAB) * m_fInverseLength);
}




// GetX
void CLineSegment::GetX(const float fU, const float fV, Vec2 *vecX) const
{
    // calculate X_x and X_y
    *vecX = m_vecA + (fU * m_vecAB) + (fV * m_vecPerpABTimesInverseLength);
}



// DistanceToPoint
float CLineSegment::DistanceToPoint(const Vec2& vecX) const
{
    // calculate u and v, then pass to the other version of this function.
    float fU, fV;
    this->GetUV(vecX, &fU, &fV);
    return this->DistanceToPoint(vecX, fU, fV);
}
float CLineSegment::DistanceToPoint(const Vec2& vecX, const float fU,
                                    const float fV) const
{
    if (fU < 0.0)
    {
        return fastsqrt((m_vecA[0]-vecX[0])*(m_vecA[0]-vecX[0])
            + (m_vecA[1]-vecX[1])*(m_vecA[1]-vecX[1]));
    }
    else if (fU > 1.0)
    {
        return fastsqrt((m_vecB[0]-vecX[0])*(m_vecB[0]-vecX[0])
            + (m_vecB[1]-vecX[1])*(m_vecB[1]-vecX[1]));
    }
    else
    {
        return fV;
    }
}



// Interpolate
void CLineSegment::Interpolate(const CLineSegment& L1, const CLineSegment& L2,
                               float fInterpFrac, Vec2 pvecEndpoints[2])
{
    // argument checking
    assert(fInterpFrac >= 0.0 && fInterpFrac <= 1.0
        && "fInterpFrac must be in the range (0..1)");

    // interpolate the two endpoints.  Increment the intepolation
    // factor slightly until the interpolated line segment has
    // non-zero length (at most two iterations).
    Vec2 vecMidA, vecMidB;
    do
    {
        // Slightly increment the interpolation factor.
        fInterpFrac += 0.000001f;

        // calculate the midpoints of the pairs of endpoints.
        vecMidA = L2.m_vecA;
        vecMidA -= L1.m_vecA;
        vecMidA *= fInterpFrac;
        vecMidA += L1.m_vecA;
        vecMidB = L2.m_vecB;
        vecMidB -= L1.m_vecB;
        vecMidB *= fInterpFrac;
        vecMidB += L1.m_vecB;
    }
    while (vecMidA == vecMidB);

    // calculate the interpolated line segment.
    Vec2 vecMidAB(vecMidB);
    vecMidAB -= vecMidA;
    
    // We want our final interpolated line segment to have its length
    // interpolated between the lengths of L1 and L2.  This will require
    // some readjustment.  Calculate the length of vecMidAB, as well
    // as the desired interpolated length.
    float fLengthActual = fastsqrt(sqrlen(vecMidAB));
    float fLength1 = L1.GetLength();
    float fLength2 = L2.GetLength();
    float fLengthDesired = fLength1 + fInterpFrac*(fLength2 - fLength1);
 
    // If the actual length differs significantly from the expected length,
    // the interpolated line segment must be adjusted.
    if ( fabs(fLengthActual - fLengthDesired) > 0.1 )
    {
        // locate the midpoint of the interpolated line segment
        Vec2 vecMidMidpoint(vecMidAB);
        vecMidMidpoint *= 0.5f;
        vecMidMidpoint += vecMidA;

        // Calculate the factor by which the length of the interpolated
        // line segment must be adjusted.
        float fAdjustFactor = fLengthDesired / fLengthActual;	

        // Stretch or shrink vecMidA and vecMidB outwards from
        // vecMidMidpoint by a factor of fAdjustFactor.
        Vec2 vecAdjustedHalfMidAB(0.5f * vecMidAB);
        vecAdjustedHalfMidAB *= fAdjustFactor;

        // Update the endpoints.
        vecMidA = vecMidMidpoint;
        vecMidA -= vecAdjustedHalfMidAB;
        vecMidB = vecMidMidpoint;
        vecMidB += vecAdjustedHalfMidAB;
    }

    // Store endpoints in the output array.
    pvecEndpoints[0] = vecMidA;
    pvecEndpoints[1] = vecMidB;
}

// Precalculate
void CLineSegment::Precalculate(const float weightP)
{
    m_vecAB = m_vecB;
    m_vecAB -= m_vecA;

    m_vecPerpAB = Vec2( m_vecAB[1], -m_vecAB[0] );

    m_fLength = static_cast<float>(len(m_vecAB));
    m_fLength2 = m_fLength * m_fLength;
    m_fLengthToTheP = static_cast<float>(pow(m_fLength, weightP));
    assert(m_fLength != 0.0f);
    if (m_fLength != 0.0f)
    {
        FP_INV_C(m_fInverseLength, m_fLength);
        m_fInverseLength2 = m_fInverseLength * m_fInverseLength;
        m_vecPerpABTimesInverseLength = m_vecPerpAB * m_fInverseLength;
    }
}
