#include "ImageMorpher.h"
#include "LineSegment.h"
#include "MorphFile.h"
#include "PixelMap.h"

#include <cstdio>
#include <cstring>
#include <string>
using std::string;

namespace
{
    // utility function to normalize pathnames to use unix-style '/'
    // directory delimters.
    void NormalizePath(char *path)
    {
        int i=0;
        while(path[i] != 0)
        {
            if (path[i] == '\\')
            {
                path[i] = '/';
            }
            ++i;
        }
    }

    const char MRF_MAGIC_STRING[4] = {'M','R','F','\0'};
    const unsigned int MRF_VERSION = 1;

    // preamble
    typedef struct
    {
        char magicString[4];
        unsigned int version;
        unsigned int dataOffset; // file offset of start of data
    } MrfPreamble;
    
    // version 1 header
    typedef struct
    {
        unsigned int filenameALength;
        unsigned int filenameBLength;
        unsigned int numLines;
        float warpFrac;
        float dissolveFrac;
        float morphParamA;
        float morphParamB;
        float morphParamP;
    } MrfHeaderV1;
}

int CMorphFile::Load(const string& filename, MorphInfo *info)
{
    FILE *inFile = fopen(filename.c_str(), "rb");
    if (inFile == NULL)
        return -1;

    char *filenameA = NULL;
    char *filenameB = NULL;
    bool done = false;
    do
    {
        unsigned int i;
        size_t count = 0;

        // Read preamble
        MrfPreamble preamble;
        count = fread(&preamble, sizeof(MrfPreamble), 1, inFile);
        if (count != 1)
            break;
        if (strncmp(preamble.magicString, MRF_MAGIC_STRING, 4) != 0)
            break;
        if (preamble.version > MRF_VERSION)
            break;

        // Read header
        MrfHeaderV1 header1;
        count = fread(&header1, sizeof(MrfHeaderV1), 1, inFile);
        if (count != 1)
            break;
        if (header1.filenameALength > 0xFFFF || header1.filenameBLength > 0xFFFF
            || header1.numLines > 0xFFFF)
            break;
        info->warpFrac     = header1.warpFrac;
        info->dissolveFrac = header1.dissolveFrac;
        info->morphParamA  = header1.morphParamA;
        info->morphParamB  = header1.morphParamB;
        info->morphParamP  = header1.morphParamP;

        // Read filename A
        fseek(inFile, preamble.dataOffset, SEEK_SET);
        filenameA = new char[header1.filenameALength];
        count = fread(filenameA, header1.filenameALength, 1, inFile);
        if (count != 1)
            break;
        NormalizePath(filenameA);
        info->filenameA = filenameA;

        // Read filename B
        filenameB = new char[header1.filenameBLength];
        count = fread(filenameB, header1.filenameBLength, 1, inFile);
        if (count != 1)
            break;
        NormalizePath(filenameB);
        info->filenameB = filenameB;

        // Read lines
        SegmentVector& linesA = *(info->linesA);
        linesA.clear();
        linesA.reserve(header1.numLines);
        SegmentVector& linesB = *(info->linesB);
        linesB.clear();
        linesB.reserve(header1.numLines);
        CLineSegment line;
        for(i=0; i<header1.numLines; ++i)
        {
            float linePair[8];
            count = fread(linePair, 8*sizeof(float), 1, inFile);
            if (count != 1)
                break;
            line.SetPoints(linePair);
            linesA.push_back(line);
            line.SetPoints(linePair+4);
            linesB.push_back(line);
        }
        if (i != header1.numLines)
            break;

        // success!
        done = true;
    }
    while(0);

    fclose(inFile);
    if (filenameA != NULL)
        delete [] filenameA;
    if (filenameB != NULL)
        delete [] filenameB;
    return done ? 0 : -1;
}


int CMorphFile::Save(const string& filename, const MorphInfo *info)
{
    FILE *outFile = fopen(filename.c_str(), "wb");
    if (outFile == NULL)
        return -1;

    bool done = false;
    do
    {
        unsigned int i;
        size_t count = 0;

        // Write preamble
        MrfPreamble preamble;
        strncpy(preamble.magicString, MRF_MAGIC_STRING, 4);
        preamble.version = MRF_VERSION;
        preamble.dataOffset = sizeof(MrfPreamble) + sizeof(MrfHeaderV1);
        count = fwrite(&preamble, sizeof(MrfPreamble), 1, outFile);
        if (count != 1)
            break;

        // Write header
        string foo;
        MrfHeaderV1 header1;
        header1.filenameALength = (unsigned int)info->imgA->GetSourceFile().length() + 1;
        header1.filenameBLength = (unsigned int)info->imgB->GetSourceFile().length() + 1;
        header1.numLines        = (unsigned int)info->linesA->size();
        header1.warpFrac        = info->warpFrac;
        header1.dissolveFrac    = info->dissolveFrac;
        header1.morphParamA     = info->morphParamA;
        header1.morphParamB     = info->morphParamB;
        header1.morphParamP     = info->morphParamP;
        count = fwrite(&header1, sizeof(MrfHeaderV1), 1, outFile);
        if (count != 1)
            break;

        // Write data
        count = fwrite(info->imgA->GetSourceFile().c_str(), header1.filenameALength, 1, outFile);
        if (count != 1)
            break;
        count = fwrite(info->imgB->GetSourceFile().c_str(), header1.filenameBLength, 1, outFile);
        if (count != 1)
            break;
        const SegmentVector& linesA = *(info->linesA);
        const SegmentVector& linesB = *(info->linesB);
        for(i=0; i<header1.numLines; ++i)
        {
            float linePair[8] =
            {
                linesA[i].GetAX(),
                linesA[i].GetAY(),
                linesA[i].GetBX(),
                linesA[i].GetBY(),
                linesB[i].GetAX(),
                linesB[i].GetAY(),
                linesB[i].GetBX(),
                linesB[i].GetBY()
            };
            count = fwrite(linePair, 8*sizeof(float), 1, outFile);
            if (count != 1)
                break;
        }
        if (i != header1.numLines)
            break;

        // success!
        done = true;
    }
    while(0);

    fclose(outFile);
    return done ? 0 : -1;
}
