#include "ImageMorpher.h"
#include "MorphView.h"
#include "LineEditor.h"
#include "ScrollableImage.h"

#pragma warning(push)
#pragma warning(disable:4311 4312)
#include <FL/Fl.H>
#include <FL/fl_ask.H>
#include <FL/fl_file_chooser.H>
#pragma warning(pop)

#if WIN32
#include <windows.h>
#endif
#include <GL/gl.h>

#include <cassert>
#include <string>
using std::string;

namespace
{
    const float WARP_FRACTION_INIT     = 0.5f;
    const float DISSOLVE_FRACTION_INIT = 0.5f;
    const float MORPH_PARAMETER_A_INIT = 0.01f;
    const float MORPH_PARAMETER_B_INIT = 2.0f;
    const float MORPH_PARAMETER_P_INIT = 0.2f;
    
    const int IMAGE_PADDING = 10;

    enum
    {
        IO_SUCCESS = 0,
        IO_FAIL,
    };
}

CMorphView::CMorphView(int x, int y, int w,int h,const char *l)
: Fl_Gl_Window(x,y,w,h,l)
{ 
    mode(FL_RGB | FL_DOUBLE);

    //m_warpFrac    = WARP_FRACTION_INIT; 
    //m_dissFrac    = DISSOLVE_FRACTION_INIT; 
    //m_morphParamA = MORPH_PARAMETER_A_INIT; 
    //m_morphParamB = MORPH_PARAMETER_B_INIT; 
    //m_morphParamP = MORPH_PARAMETER_P_INIT; 

    m_currentImageId = IMG_OUT;
    m_isActive = true;
    m_scrollImgA = new ScrollableImage;
    m_scrollImgB = new ScrollableImage;
    m_scrollImgC = new ScrollableImage;
    m_lineEditor = new CLineEditor;
    m_currentToolId = TOOL_ADD;
    m_mouseClickX = m_mouseClickY = m_mouseLastX = m_mouseLastY = 0;
}

CMorphView::~CMorphView(void)
{
    if (m_lineEditor != NULL)
        delete m_lineEditor;
    if (m_scrollImgA != NULL)
        delete m_scrollImgA;
    if (m_scrollImgB != NULL)
        delete m_scrollImgB;
    if (m_scrollImgC != NULL)
        delete m_scrollImgC;
}

CMorphView::ImageID CMorphView::GetWhichImage(int x,int y) const
{ 
    int windowWidth = w();
    int imageWidth = windowWidth/3;
    if (x > IMAGE_PADDING && x < imageWidth-IMAGE_PADDING)
        return IMG_A;
    if (x > imageWidth+IMAGE_PADDING && x < 2*imageWidth-IMAGE_PADDING)
        return IMG_B;
    if (x > 2*imageWidth+IMAGE_PADDING && x < 3*imageWidth-IMAGE_PADDING)
        return IMG_OUT;
    return IMG_OUT;  // should be NONE
}

ScrollableImage *CMorphView::GetScrollImage(const ImageID whichImage) const
{
    switch(whichImage)
    {
    case IMG_A:
        return m_scrollImgA;
    case IMG_B:
        return m_scrollImgB;
    case IMG_OUT:
        return m_scrollImgC;
    default:
        assert(0);
        return NULL; // should NEVER happen
    }
}


int CMorphView::OpenImage(const ImageID whichImage, const string& filename)
{
    CPixelMap *img = NULL;
    ScrollableImage *scrollImg = NULL;
    int retval = IO_FAIL;
    switch(whichImage)
    {
    case IMG_A:
        img = &m_imgA;
        scrollImg = m_scrollImgA;
        break;
    case IMG_B:
        img = &m_imgB;
        scrollImg = m_scrollImgB;
        break;
    case IMG_OUT:
        img = &m_imgC;
        scrollImg = m_scrollImgC;
        break;
    }
    if (img == NULL)
    {
        fl_alert("Attempt to load into invalid image (should never happen!)");
    }
    else if (img->Load(filename) != 0)
    {
        fl_alert("Could not load image %s", filename.c_str());
    }
    else
    {
        // success!
        scrollImg->SetImageSize(img->GetWidth(), img->GetHeight());
        scrollImg->ZoomPreset(ScrollableImage::ZOOM_FIT_VISIBLE);
        retval = IO_SUCCESS;
    }
    redraw();
    return retval;
}


int CMorphView::SaveImage(const ImageID whichImage, const string& filename)
{
    CPixelMap *img = NULL;
    int retval = IO_FAIL;
    switch(whichImage)
    {
    case IMG_A:
        img = &m_imgA;
        break;
    case IMG_B:
        img = &m_imgB;
        break;
    case IMG_OUT:
        img = &m_imgC;
        break;
    }
    if (img == NULL || img->GetPixels() == NULL)
    {
        fl_alert("Attempt to save to invalid image (should never happen!)");
    }
    else if (img->Save(filename, 0.75f) != 0)
    {
        fl_alert("Error saving image file %s", filename.c_str());
    }
    else
    {
        // success!
        retval = IO_SUCCESS;
    }
    return retval;
}


int CMorphView::GetImageDimensions(const ImageID whichImage, int *outWidth,
                                   int *outHeight) const
{
    const CPixelMap *img = NULL;
    int retval = IO_FAIL;
    switch(whichImage)
    {
    case IMG_A:
        img = &m_imgA;
        break;
    case IMG_B:
        img = &m_imgB;
        break;
    case IMG_OUT:
        img = &m_imgC;
        break;
    }
    if (img == NULL)
        return IO_FAIL;
    
    if (outWidth != NULL)
        *outWidth = img->GetWidth();
    if (outHeight != NULL)
        *outHeight = img->GetHeight();
    return IO_SUCCESS;
}


bool CMorphView::IsReadyToMorph(void) const
{
    if (m_imgA.GetPixels() == NULL || m_imgB.GetPixels() == NULL)
        return false;
    return true;
}


void CMorphView::GatherMorphInfo(MorphInfo *info)
{
    info->imgA = &m_imgA;
    info->imgB = &m_imgB;
    info->imgOut = const_cast<CPixelMap*>(&m_imgC);
    m_lineEditor->GatherMorphInfo(info); // line editor has some info too!
}


void CMorphView::DisperseMorphInfo(const MorphInfo *info)
{
    const char *filename;
    int retval;

    // Load image A
    filename = info->filenameA.c_str();
    retval = -1;
    while(true)
    {
        retval = OpenImage(IMG_A, filename);
        if (retval == 0)
            break; // success!
        int browse = fl_ask("Unable to load image %s. Search for it manually?",
            filename);
        if (!browse)
            break; // user didn't want to browse
        filename = fl_file_chooser("Select an image", "*.{jpg,png,bmp,tga,gif}",
            filename);
        if (filename == NULL)
            break; // user cancelled browsing
    }
    if (retval != 0)
    {
        fl_alert("Error loading MRF file: image loading failure.");
        return; // can't continue
    }

    // Load image B
    filename = info->filenameB.c_str();
    retval = -1;
    while(true)
    {
        retval = OpenImage(IMG_B, filename);
        if (retval == 0)
            break; // success!
        int browse = fl_ask("Unable to load image %s. Search for it manually?",
            filename);
        if (!browse)
            break; // user didn't want to browse
        filename = fl_file_chooser("Select an image", "*.{jpg,png,bmp,tga,gif}",
            filename);
        if (filename == NULL)
            break; // user cancelled browsing
    }
    if (retval != 0)
    {
        fl_alert("Error loading MRF file: image loading failure.");
        return; // can't continue
    }
    
    // Remove anything lingering in output image.
    m_imgC.Unload();

    m_lineEditor->DisperseMorphInfo(info);
}


void CMorphView::DrawImageAndLines(const ImageID whichImage) const
{
    // Get a pointer to the appropriate image
    const CPixelMap *img = NULL;
    switch(whichImage)
    {
    case IMG_A:
        img = &m_imgA;
        break;
    case IMG_B:
        img = &m_imgB;
        break;
    case IMG_OUT:
        img = &m_imgC;
        break;
    }
    if (img == NULL || img->GetPixels() == NULL)
        return;

    int minX, minY, maxX, maxY;
    ScrollableImage *scrollImg = GetScrollImage(whichImage);

    // Update the scollable image's dimensions.  This is really only
    // necessary for the output image.
    scrollImg->SetImageSize(img->GetWidth(), img->GetHeight());

    // Set up the scissor rectangle for this image
    scrollImg->GetViewportRect(&minX, &minY, &maxX, &maxY);
    glScissor(minX, minY, maxX-minX, maxY-minY);

    // Get the rectangle of the current image, with zoom & scroll
    // applied, in window coordinates.
    scrollImg->GetImageRect(&minX, &minY, &maxX, &maxY);

    // store current pixel zoom level, and then apply the correct zoom
    // for this image.  We have to flip the y coordinates, regrettably,
    // due to a disagreement between FLTK and OpenGL as to which way is up.
    float zoomX, zoomY;
    glGetFloatv(GL_ZOOM_X, &zoomX);
    glGetFloatv(GL_ZOOM_Y, &zoomY);
    glPixelZoom((float)(maxX-minX) / (float)img->GetWidth(),
        -1.0f * (float)(maxY-minY) / (float)img->GetHeight());

    // Draw the image.  Use the glBitmap() trick to effectively set the
    // raster position offscreen without it being invalidated.
    glRasterPos2i(0,0);
    glBitmap(0, 0, 0, 0, (float)minX, (float)-minY, NULL);
    glDrawPixels(img->GetWidth(), img->GetHeight(), GL_RGBA, GL_UNSIGNED_BYTE,
        img->GetPixels());

    // restore old pixel zoom
    glPixelZoom(zoomX, zoomY);

    // Draw the correspondence lines associated with this image.
    m_lineEditor->DrawLines(whichImage, Vec2((Real)minX, (Real)minY),
        Vec2((Real)maxX, (Real)maxY));
}


void CMorphView::draw()  
{ 
    // the valid() property may be used to avoid reinitializing your 
    // GL transformation for each redraw: 
    if (!valid())
    { 
        valid(1); 
        glMatrixMode(GL_PROJECTION); 
        glLoadIdentity(); 
        // Specify window size 
        glViewport(0,0,w(),h()); 
        // Specify coordinate mapping 
        glOrtho(0, w(), 0,h(),-1.0,1.0); 
        glMatrixMode(GL_MODELVIEW); 
        glDrawBuffer(GL_FRONT_AND_BACK); 
        glDisable(GL_DEPTH_TEST); 
        glDisable(GL_LIGHTING); 
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glDisable(GL_CULL_FACE);

    }

    // Put the origin in the upper-left corner.
    glPushMatrix();
    glLoadIdentity();
    glTranslatef(0.0f, (float)h(), 0.0f);
    glScalef(1.0f, -1.0f, 1.0f);

    // update scrolling image state
    int imageViewportWidth = w() / 3;
    int imageViewportHeight = h();
    m_scrollImgA->SetViewportOrigin(IMAGE_PADDING,
        IMAGE_PADDING);
    m_scrollImgA->SetViewportSize(imageViewportWidth - 2*IMAGE_PADDING,
        imageViewportHeight - 2*IMAGE_PADDING);
    
    m_scrollImgB->SetViewportOrigin(imageViewportWidth + IMAGE_PADDING,
        IMAGE_PADDING);
    m_scrollImgB->SetViewportSize(imageViewportWidth - 2*IMAGE_PADDING,
        imageViewportHeight - 2*IMAGE_PADDING);
    
    m_scrollImgC->SetViewportOrigin(2*imageViewportWidth + IMAGE_PADDING,
        IMAGE_PADDING);
    m_scrollImgC->SetViewportSize(imageViewportWidth - 2*IMAGE_PADDING,
        imageViewportHeight - 2*IMAGE_PADDING);

    // Record the current GL state state, so we can restore it later.
    bool isScissorEnabled = glIsEnabled(GL_SCISSOR_TEST) != 0;
    int oldBox[4];
    if (isScissorEnabled)
        glGetIntegerv(GL_SCISSOR_BOX, oldBox);

    // draw images and lines
    glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_SCISSOR_TEST);
    DrawImageAndLines(IMG_A);
    DrawImageAndLines(IMG_B);
    DrawImageAndLines(IMG_OUT);

    // Restore original scissor rect settings.
    if (isScissorEnabled)
        glScissor(oldBox[0], oldBox[1], oldBox[2], oldBox[3]);
    else
        glDisable(GL_SCISSOR_TEST);

    glPopMatrix();
    glDrawBuffer(GL_BACK);
    glFlush();
}


int CMorphView::HandleMousePush()
{
    int handled = 0;
    int imgx = Fl::event_x(); 
    int imgy = Fl::event_y(); 
    float cordx,cordy;
    
    // Figure out which image the user clicked in.
    m_currentImageId = GetWhichImage(imgx, imgy); 
    GetScrollImage(m_currentImageId)->PixelToImage(imgx, imgy, &cordx, &cordy);

    int buttonState = Fl::event_state();
    if ((buttonState & FL_BUTTON1) != 0)
    {
        // left mouse button presses are handled by the line editor.
        // Regrettably, we must invert the y coordinates so FLTK and OpenGL
        // play nicely.
        handled = m_lineEditor->HandleMousePush(m_currentImageId,
            m_currentToolId, cordx, 1.0f-cordy);
    }
    else if ((buttonState & FL_BUTTON3) != 0)
    {
        // right mouse button events are always scrolls/zooms.
        m_mouseClickX = m_mouseLastX = imgx;
        m_mouseClickY = m_mouseLastY = imgy;
        handled = 1;
    }

    redraw();
    return handled;
} 

int CMorphView::HandleMouseRelease()
{
    int handled = 0;
    int imgx = Fl::event_x(); 
    int imgy = Fl::event_y();
    float cordx,cordy; 
    
    if (m_currentImageId != IMG_OUT)
    {
        GetScrollImage(m_currentImageId)->PixelToImage(imgx, imgy,
            &cordx, &cordy);
    }

    int buttonState = Fl::event_state();
    if ((buttonState & FL_BUTTON1) != 0)
    {
        // left mouse button presses are handled by the line editor.
        // Regrettably, we must invert the y coordinates so FLTK and OpenGL
        // play nicely.
        handled = m_lineEditor->HandleMouseRelease(m_currentImageId,
            m_currentToolId, cordx, 1.0f-cordy);
    }
    else if ((buttonState & FL_BUTTON3) != 0)
    {
        // nothing to do for zoom/scroll when button is released.
        handled = 1;
    }
    
    redraw();
    return handled;
} 


int CMorphView::HandleMouseDrag()
{
    int handled = 0;
    int imgx = Fl::event_x(); 
    int imgy = Fl::event_y();
    float cordx,cordy;
    
    if (m_currentImageId != IMG_OUT)
    {
        GetScrollImage(m_currentImageId)->PixelToImage(imgx, imgy,
            &cordx, &cordy); 
    }

    int buttonState = Fl::event_state();
    if ((buttonState & FL_BUTTON1) != 0)
    {
        // left mouse button presses are handled by the line editor.
        // Regrettably, we must invert the y coordinates so FLTK and OpenGL
        // play nicely.
        handled = m_lineEditor->HandleMouseDrag(m_currentImageId,
            m_currentToolId, cordx, 1.0f-cordy);
    }
    else if ((buttonState & FL_BUTTON3) != 0)
    {
        // right mouse button events are always zoom/scroll.
        if ((buttonState & FL_SHIFT) == 0)
        {
            // shift is not held; we're scrolling.

            GetScrollImage(m_currentImageId)->Scroll(imgx - m_mouseLastX,
                imgy - m_mouseLastY);
        }
        else
        {
            // shift is held; we're zooming.
            const float dZoom = -0.01f;
            GetScrollImage(m_currentImageId)->Zoom(dZoom*(imgy-m_mouseLastY),
                m_mouseClickX, m_mouseClickY);
        }
        m_mouseLastX = imgx;
        m_mouseLastY = imgy;
        handled = 1;
    }
    
    redraw();
    return handled;
}


int CMorphView::HandleMouseWheel(void)
{
    int handled = 0;
    int imgx  = Fl::event_x(); 
    int imgy  = Fl::event_y();
    int wheel_dx = Fl::event_dx();
    int wheel_dy = Fl::event_dy();
    
    redraw();
    return handled;
}


int CMorphView::handle(int event)  
{
    if (m_isActive)
    {
        switch(event)
        {
        case FL_PUSH:
            return HandleMousePush(); 
        case FL_DRAG:
            return HandleMouseDrag(); 
        case FL_RELEASE:
            return HandleMouseRelease(); 
        case FL_MOUSEWHEEL:
            return HandleMouseWheel();
        default:
            return Fl_Gl_Window::handle(event);
        } 
    }
    return 0;
}
