#include "PixelMap.h"

#include "ilwrap.h"

#if !defined(IGNORE_DEPEND)
#include <cassert>
#include <cstdlib>
#include <iostream>
using std::cerr;
using std::endl;
#include <string>
using std::string;
#endif // IGNORE_DEPEND


// These values are copied from http://openil.sourceforge.net/features.php
const char * const CPixelMap::LOAD_SUFFIXES = "*.{bmp,cut,dcx,dds,ico,jpg,lbm,lif,mdl,pcd,pcx,pic,png,pnm,psd,psp,raw,sgi,tga,tif,wal,act,pal}";
const char * const CPixelMap::SAVE_SUFFIXES = "*.{bmp,dds,jpg,pcx,png,pnm,raw,sgi,tga,tif,pal}";


///////////////////////////////////////////////////////////////////////////////
// Construction/Destruction
///////////////////////////////////////////////////////////////////////////////

CPixelMap::CPixelMap(void) : m_nWidth(0), m_nHeight(0), m_nBytesPerPixel(0),
	m_pPixels(NULL), m_sSourceFile("undefined")
{
}

CPixelMap::CPixelMap(const CPixelMap& rhs)
{
  m_nWidth = rhs.m_nWidth;
  m_nHeight = rhs.m_nHeight;
  m_nBytesPerPixel = rhs.m_nBytesPerPixel;
  m_pPixels = new Pixel8[m_nWidth * m_nHeight * m_nBytesPerPixel];
  assert(m_pPixels != NULL);
}

CPixelMap::~CPixelMap(void)
{
  if (m_pPixels != NULL)
    {
      delete [] m_pPixels;
      m_pPixels = NULL;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Overloaded Operators
///////////////////////////////////////////////////////////////////////////////

const CPixelMap& CPixelMap::operator=(const CPixelMap& rhs)
{
    // test for self-assignment
    if (this == &rhs)
    {
        return *this;
    }

    Initialize(rhs.m_nWidth, rhs.m_nHeight, rhs.m_nBytesPerPixel);
    memcpy(m_pPixels, rhs.m_pPixels, m_nWidth*m_nHeight*m_nBytesPerPixel);

    return *this;
}



///////////////////////////////////////////////////////////////////////////////
// Public Methods
///////////////////////////////////////////////////////////////////////////////

int CPixelMap::Load(const string &sFilename)
{
    int retval = ILWrapLoadImage(sFilename, *this);
    if (retval == 0)
        m_sSourceFile = sFilename;
    return retval;
}

// Save
int CPixelMap::Save(const string& sFilename, const float fQuality)
{
    int retval = ILWrapSaveImage(sFilename, *this);
    if (retval == 0)
        m_sSourceFile = sFilename;
    return retval;
}


// Unload
void CPixelMap::Unload(void)
{
    if (m_pPixels != NULL)
    {
        delete [] m_pPixels;
        m_pPixels = NULL;
    }
    m_nWidth = m_nHeight = m_nBytesPerPixel = 0;
    m_sSourceFile = "undefined";
}



// Initialize
void CPixelMap::Initialize(const unsigned int unW, const unsigned int unH,
	const unsigned int unBytesPerPixel)
{
	assert(unBytesPerPixel >= 1 && unBytesPerPixel <= 4); 
	if (m_pPixels != NULL)
	{
		delete [] m_pPixels;
		m_pPixels = NULL;
	}

	m_pPixels = new Pixel8[unW*unH*unBytesPerPixel];
	assert(m_pPixels != NULL);
	m_nWidth = unW;
	m_nHeight = unH;
	m_nBytesPerPixel = unBytesPerPixel;
    m_sSourceFile = "undefined";
}



// GetPixel
Pixel32 CPixelMap::GetPixel(int nX, int nY) const
{
	// check boundaries
	if (nX < 0 || nX >= m_nWidth || nY < 0 || nY > m_nHeight)
	{
		Pixel32 dummy = {0,0,0,0};
		cerr << "ERROR: pixel [" << nX << "," << nY << "] is out of range.\n";
		return dummy;
	}

	// Grab the correct pixel
	switch(m_nBytesPerPixel)
	{
	case 1:
		{
			Pixel8 pix = GetPixelFast1(nX, nY);
			Pixel32 retPix = {pix,pix,pix,pix};
			return retPix;
		}
	case 3:
		{
			assert(sizeof(Pixel24) == 3); // or else GetPixelFast will return junk
			Pixel24 pix = GetPixelFast3(nX, nY);
			Pixel32 retPix = {pix.r,pix.g,pix.b,255};
			return retPix;
		}
	case 4:
		return GetPixelFast4(nX, nY);
	default:
		{
			Pixel32 dummy = {0,0,0,0};
			cerr << "ERROR: invalid bit depth in GetPixel()\n";
			return dummy;
		}
	}
}
