// A platform-independent interface class for working with image files in
// memory.

#if !defined (PIXELMAP_H)
#define PIXELMAP_H

#if !defined(IGNORE_DEPEND)
#include <string>
#endif // IGNORE_DEPEND

// pixel type definitions
typedef unsigned char Pixel8;
typedef struct
{
	Pixel8 r,g,b;
} Pixel24;
typedef struct
{
	Pixel8 r,g,b,a;	
} Pixel32;

class CPixelMap
{
public:
	CPixelMap(void);
	CPixelMap(const CPixelMap& rhs);
	virtual ~CPixelMap();

	const CPixelMap& operator=(const CPixelMap& rhs);

	// Load an image file.  Returns zero if successful.
	int Load(const std::string& sFilename);

	// Save the image file.  The file type is determined automatically
	// from the filename suffix.  The quality parameter is applied wherever
	// appropriate, and has a range from 0.0 to 1.0. Returns zero if
	// successful, nonzero on an error.
	int Save(const std::string& sFilename, const float fQuality = 0.5f);

    // Unloads whatever pixel data is loaded.  The image is essentially emptied.
    void Unload(void);

	// query the image file
	int GetWidth(void) const { return m_nWidth; }
	int GetHeight(void) const { return m_nHeight; }
	int GetBytesPerPixel(void) const { return m_nBytesPerPixel; }
	Pixel8* GetPixels(void) const { return m_pPixels; }
	const std::string& GetSourceFile(void) const { return m_sSourceFile; }

	//! Re-initializes the image with the given dimensions and color depth.  Any
	// existing piel data will be erased.  The new pixel data will be uninitialized.
	void Initialize(const unsigned int unW, const unsigned int unH,
		const unsigned int unBytesPerPixel = 4);

	// Get the value of a specific pixel.  Two versions: GetPixel performs
	// error-checking, GetPixelFast* does not.
	Pixel32 GetPixel(int nX, int nY) const;
	Pixel8 GetPixelFast1(int nX, int nY) const
	{
		return m_pPixels[nY*m_nWidth + nX];
	}
	Pixel24 GetPixelFast3(int nX, int nY) const
	{
		return ((Pixel24*)(m_pPixels))[nY*m_nWidth + nX];
	}
	Pixel32 GetPixelFast4(int nX, int nY) const
	{
		return ((Pixel32*)(m_pPixels))[nY*m_nWidth + nX];
	}

    void SetPixel4(const int nX, const int nY, const Pixel8 color[3])
    {
        if (nX < 0 || nX >= m_nWidth || nY < 0 || nY >= m_nHeight)
            return;
        Pixel8 *pixel = m_pPixels + 4*(nY*m_nWidth + nX);
        *pixel++ = color[2];
        *pixel++ = color[1];
        *pixel   = color[0];
    }

    //! Lists of legal suffixes for filetypes that this class can load and
    //! save.  Takes the form "*.{aaa,bbb,ccc}", suitable for passing to
    //! fl_file_chooser().
    static const char * const LOAD_SUFFIXES;
    static const char * const SAVE_SUFFIXES;
    
private:
	//! Image attributes.
	int m_nWidth, m_nHeight, m_nBytesPerPixel;

	//! Image pixels.
	Pixel8 *m_pPixels;

	//! Image source (the file from which it was loaded -- "undefined"
	// if undefined.
	std::string m_sSourceFile;
};

#endif // PIXELMAP_H
