#include "ScrollableImage.h"

#include <cassert>
#include <cstdio>

namespace
{

}

ScrollableImage::ScrollableImage(void)
{
    SetViewportOrigin(0,0);
    SetViewportSize(1,1);
    SetImageOrigin(0,0);
    SetImageSize(1,1);
    SetImageScaleBounds(0.2f, 5.0f);
    SetImageScale(1.0f);
}

ScrollableImage::~ScrollableImage(void)
{

}


void ScrollableImage::SetViewportOrigin(int originX, int originY)
{
    m_viewOriginX = originX;
    m_viewOriginY = originY;
}
void ScrollableImage::SetViewportSize(int sizeX, int sizeY)
{
    assert(sizeX > 0 && sizeY > 0);
    if (sizeX <= 0 || sizeY <= 0)
        return;
    m_viewSizeX = sizeX;
    m_viewSizeY = sizeY;
}
void ScrollableImage::SetImageOrigin(int originX, int originY)
{
    m_imgOriginX = originX;
    m_imgOriginY = originY;
    // make sure the new origin is within the legal bounds
    ConstrainImageOrigin();
}
void ScrollableImage::SetImageSize(int sizeX, int sizeY)
{
    assert(sizeX > 0 && sizeY > 0);
    if (sizeX <= 0 || sizeY <= 0)
        return;
    m_imgSizeX = sizeX;
    m_imgSizeY = sizeY;
}
void ScrollableImage::SetImageScale(float scale)
{
    if (scale < m_imgScaleMin)
        m_imgScale = m_imgScaleMin;
    else if (scale > m_imgScaleMax)
        m_imgScale = m_imgScaleMax;
    else
        m_imgScale = scale;
}
void ScrollableImage::SetImageScaleBounds(float minScale, float maxScale)
{
    assert(minScale > 0.0f && maxScale > 0.0f && maxScale >= minScale);
    if (minScale <= 0.0f || maxScale <= 0.0f && maxScale < minScale)
        return;
    m_imgScaleMin = minScale;
    m_imgScaleMax = maxScale;
}

void ScrollableImage::GetViewportRect(int *minX, int *minY,
                                      int *maxX, int *maxY) const
{
    if (minX != NULL)
        *minX = m_viewOriginX;
    if (minY != NULL)
        *minY = m_viewOriginY;
    if (maxX != NULL)
        *maxX = m_viewOriginX + m_viewSizeX;
    if (maxY != NULL)
        *maxY = m_viewOriginY + m_viewSizeY;
}
void ScrollableImage::GetImageRect(int *minX, int *minY,
                                   int *maxX, int *maxY) const
{
    if (minX != NULL)
        *minX = m_viewOriginX + m_imgOriginX;
    if (minY != NULL)
        *minY = m_viewOriginY + m_imgOriginY;
    if (maxX != NULL)
        *maxX = m_viewOriginX + m_imgOriginX + (int)(m_imgSizeX * m_imgScale);
    if (maxY != NULL)
        *maxY = m_viewOriginY + m_imgOriginY + (int)(m_imgSizeY * m_imgScale);
}


void ScrollableImage::Scroll(int scrollX, int scrollY)
{
    m_imgOriginX += scrollX;
    m_imgOriginY += scrollY;
    ConstrainImageOrigin();
}

void ScrollableImage::Zoom(float deltaScale, int centerX, int centerY)
{
    // convert pixel coordinates to image coordinates before we update the
    // scale!
    float imgX, imgY;
    PixelToImage(centerX, centerY, &imgX, &imgY);

    float oldScale = m_imgScale;
    SetImageScale(m_imgScale + deltaScale);
    // update the actual delta scale
    deltaScale = m_imgScale - oldScale;

    // Adjust the image origin such that the centerX and centerY still
    // correspond to the same image coordinates.
    SetImageOrigin(m_imgOriginX - (int)(deltaScale*imgX*m_imgSizeX),
        m_imgOriginY - (int)(deltaScale*imgY*m_imgSizeY));
}

void ScrollableImage::ZoomPreset(ZoomLevel level)
{
    float fitWidthScale  = (float)m_viewSizeX / (float)m_imgSizeX;
    float fitHeightScale = (float)m_viewSizeY / (float)m_imgSizeY;
    switch(level)
    {
    case ZOOM_ACTUAL_SIZE:
        SetImageScale(1.0f);
        SetImageOrigin(0,0);
        break;
    case ZOOM_FIT_VISIBLE:
        SetImageScale( (fitWidthScale < fitHeightScale) ?
            fitWidthScale : fitHeightScale );
        SetImageOrigin(0,0);
        break;
    case ZOOM_FIT_WIDTH:
        SetImageScale(fitWidthScale);
        SetImageOrigin(0,0);
        break;
    case ZOOM_FIT_HEIGHT:
        SetImageScale(fitHeightScale);
        SetImageOrigin(0,0);
        break;
    default:
        break;
    }
}


void ScrollableImage::PixelToImage(int px, int py, float *ix, float *iy) const
{
    if (ix != NULL)
        *ix = (float)(px - m_viewOriginX - m_imgOriginX) / (float)(m_imgSizeX*m_imgScale);
    if (iy != NULL)
        *iy = (float)(py - m_viewOriginY - m_imgOriginY) / (float)(m_imgSizeY*m_imgScale);
}
void ScrollableImage::ImageToPixel(float ix, float iy, int *px, int *py) const
{
    if (px != NULL)
        *px = (int)((ix * m_imgSizeX * m_imgScale) + m_viewOriginX + m_imgOriginX);
    if (py != NULL)
        *py = (int)((iy * m_imgSizeY * m_imgScale) + m_viewOriginY + m_imgOriginY);
}   


void ScrollableImage::ConstrainImageOrigin(void)
{
    // One of two rules is applied, depending on whether the image currently
    // fits completely inside the viewport.
    int scaledImgSizeX = (int)(m_imgSizeX * m_imgScale);
    int viewMaxX = m_viewSizeX + 1;
    if (scaledImgSizeX <= m_viewSizeX)
    {
        // If the scaled image fits in the window, ensure that the image hasn't
        // scrolled off the screen in this dimension.
        if (m_imgOriginX < 0)
            m_imgOriginX = 0;
        else if (m_imgOriginX + scaledImgSizeX > viewMaxX)
            m_imgOriginX = viewMaxX - scaledImgSizeX;
    }
    else
    {
        // If the scaled image is too big to fit on-screen in this dimension,
        // ensure that we're viewing as much of the image as possible.
        if (m_imgOriginX > 0)
            m_imgOriginX = 0;
        else if (m_imgOriginX + scaledImgSizeX < viewMaxX)
            m_imgOriginX = viewMaxX - scaledImgSizeX;
    }

    // Repeat the same constraints in the Y dimension
    int scaledImgSizeY = (int)(m_imgSizeY * m_imgScale);
    int viewMaxY = m_viewSizeY + 1;
    if (scaledImgSizeY <= m_viewSizeY)
    {
        // If the scaled image fits in the window, ensure that the image hasn't
        // scrolled off the screen in this dimension.
        if (m_imgOriginY < 0)
            m_imgOriginY = 0;
        else if (m_imgOriginY + scaledImgSizeY > viewMaxY)
            m_imgOriginY = viewMaxY - scaledImgSizeY;
    }
    else
    {
        // If the scaled image is too big to fit on-screen in this dimension,
        // ensure that we're viewing as much of the image as possible.
        if (m_imgOriginY > 0)
            m_imgOriginY = 0;
        else if (m_imgOriginY + scaledImgSizeY < viewMaxY)
            m_imgOriginY = viewMaxY - scaledImgSizeY;
    }
}
