#include "bilinear.h"
#include "fastmath.h"

#include <cmath>
#include <cassert>


// BILINEAR -- takes coordinates x,y in floating-point format and uses
// bilinear interpolation to determine the color at that point in the
// image.  The interpolated color values are stored in the rgb array.
// the return value is zero if nothing goes wrong, anything else if
// there was an error.
void bilinear(const CPixelMap& img, float x, float y, unsigned char rgb[3])
{
    // make sure the pointer arguments aren't void
    assert(img.GetBytesPerPixel() == 4);

    // find closest coordinates.  do bounds-checking
    int x_int, y_int;
    FloatToInt(&x_int, (float)floor(x));
    FloatToInt(&y_int, (float)floor(y));
    int nx = img.GetWidth(), ny = img.GetHeight(), nx4 = 4*nx;  

    // if the coordinates are out of bounds, normalize them back to the
    // edges of the image.  This is NOT an error!
    if (x_int < 0)
    {
        x_int = 0;
        x = 0.0f;
    }
    else if (x_int >= nx-1)
    {
        x_int = nx-2;
        x = (float)(x_int);
    }
    if (y_int < 0)
    {
        y_int = 0;
        y = 0.0f;
    }
    else if (y_int >= ny-1)
    {
        y_int = ny-2;
        y = (float)(y_int);
    }

    // having extracted the integer half of x and y, convert them to
    // the range 0..1.  WITHOUT an fmod!
    x = x - x_int;
    y = y - y_int;

    // get points to the surrounding pixels in img->pix, for convenience
    unsigned char *imgpix = (unsigned char *)(img.GetPixels());

    unsigned char *P1 = imgpix + 4*(y_int*nx + x_int);
    unsigned char *P2 = P1 + 4;
    unsigned char *P3 = P1 + nx4;
    unsigned char *P4 = P2 + nx4;

    // calculate the final, interpolated color values
    int finalRed, finalGreen, finalBlue;
    float temp1_r, temp2_r, temp1_g, temp2_g, temp1_b, temp2_b;

    temp1_r = P1[0] + x*(P2[0]-P1[0]);
    temp1_g = P1[1] + x*(P2[1]-P1[1]);
    temp1_b = P1[2] + x*(P2[2]-P1[2]);
    temp2_r = P3[0] + x*(P4[0]-P3[0]);
    temp2_g = P3[1] + x*(P4[1]-P3[1]);
    temp2_b = P3[2] + x*(P4[2]-P3[2]);

    FloatToInt(&finalRed, temp1_r + y*(temp2_r-temp1_r));
    assert( finalRed <= 255 ); assert( finalRed >= 0 );
    rgb[0] = (unsigned char)finalRed;

    FloatToInt(&finalGreen, temp1_g + y*(temp2_g-temp1_g));
    assert( finalGreen <= 255 ); assert( finalGreen >= 0 );
    rgb[1] = (unsigned char)finalGreen;

    FloatToInt(&finalBlue, temp1_b + y*(temp2_b-temp1_b));
    assert( finalBlue <= 255 ); assert( finalBlue >= 0 );
    rgb[2] = (unsigned char)finalBlue;
}
