#include "fastmath.h"  // fast math routines
#include "ImageMorpher.h"
#include "MorphFile.h"
#include "PixelMap.h"

#include <revel.h>

#include <cstdlib> // exit, strtod, random
#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
#include <string>
using std::string;

namespace
{
    bool movieMode = false;
    int numFrames = 60;
    float frameRate = 30.0f;
    float quality = 1.0f;
    CPixelMap imgA, imgB, imgOut;
    SegmentVector linesA, linesB;
    string filenameOut, filenameMrf;
}

// USAGE() -- prints usage information
void usage(void)
{
    cerr << "usage: kmorph mymorph.mrf <OUTPUT FILE> [OPTIONS]" << endl
        << endl
        << "GLOBAL OPTIONS:" << endl
        << "-x outputWidth     Override the width of the output image/movie" << endl
        << "-y outputHeight    Override the height of the output image/movie" << endl
        << endl
        << "MOVIE OPTIONS:" << endl
        << "-m                 Output a movie instead of a still image" << endl
        << "-n numFrames       In movie mode, specify the number of frames to render" << endl
        << "-r frameRate       In movie mode, specify the frame rate" << endl
        << "-q quality         Override the quality of the movie [0..1]" << endl
        << endl
        << "SINGLE-FRAME OPTIONS:" << endl
        << "-w warpFrac        Override the warp value [0..1]" << endl
        << "-d dissolveFrac    Override the dissolve value [0..1]" << endl
        << "-a paramA          Override the morph paramater A" << endl
        << "-b paramA          Override the morph paramater B" << endl
        << "-p paramA          Override the morph paramater P" << endl
        << endl
        << "Examples:" << endl
        << "    kmorph AB.mrf AB.jpg -w 0.5 -d 0.5" << endl
        << "    kmorph AB.mrf AB.avi -m -n 60 -r 30" << endl;
}

// PROCESS_ARGS -- processes and tests the command line arguments
int process_args(int argc, char *argv[], MorphInfo &info)
{
    if (argc < 3 || argc > 13)
    {
        usage();
        exit(1);
    }

    // process required arguments
    filenameMrf = argv[1];
    filenameOut = argv[2];

    int error = 0;
    if (!error)
    {
        info.linesA = &linesA;
        info.linesB = &linesB;
        error = CMorphFile::Load(filenameMrf, &info);
    }
    if (error)
    {
        cerr << "Error loading morph file " << filenameMrf << endl;
        return -1;
    }

    // Load the images
    info.imgA = &imgA;
    info.imgB = &imgB;
    info.imgOut = &imgOut;
    char filename[256];
    int retval;

    // Load image A
    strncpy(filename, info.filenameA.c_str(), 256);
    retval = -1;
    while(true)
    {
        filename[255] = 0;
        retval = imgA.Load(filename);
        if (retval == 0)
            break; // success!
        printf("Unable to locate image %s. Enter path to image: ", filename);
    	scanf("%s", filename);
    }

    // Load image B
    strncpy(filename, info.filenameB.c_str(), 256);
    retval = -1;
    while(true)
    {
        filename[255] = 0;
        retval = imgB.Load(filename);
        if (retval == 0)
            break; // success!
        printf("Unable to locate image %s. Enter path to image: ", filename);
	    scanf("%s", filename);
    }
    info.outWidth  = imgB.GetWidth();
    info.outHeight = imgB.GetHeight();

    // process optional arguments
    int i;
    for(i=3; i < argc; ++i)
    {
        if (strcmp(argv[i], "-m") == 0)
        {
            movieMode = true;
        }
        else if (strcmp(argv[i], "-n") == 0 && i < argc-1)
        {
            int temp = atoi(argv[++i]);
            if (temp <= 1)
                cerr << "WARNING: Ignoring numFrames (must be >1)\n";
            else
                numFrames = temp;
        }
        else if (strcmp(argv[i], "-r") == 0 && i < argc-1)
        {
            float temp = (float)atof(argv[++i]);
            if (temp <= 0.0f)
                cerr << "WARNING: Ignoring frameRate (must be >0)\n";
            else
                frameRate = temp;
        }
        else if (strcmp(argv[i], "-q") == 0 && i < argc-1)
        {
            float temp = (float)atof(argv[++i]);
            if (temp < 0.0f || temp > 1.0f)
                cerr << "WARNING: Ignoring quality (out of range [0..1])\n";
            else
                quality = temp;
        }
        else if (strcmp(argv[i], "-w") == 0 && i < argc-1)
        {
            float warpFrac = (float)atof(argv[++i]);
            if (warpFrac < 0.0f || warpFrac > 1.0f)
                cerr << "WARNING: Ignoring warpFrac (out of range [0..1])\n";
            else
                info.warpFrac = warpFrac;
        }
        else if (strcmp(argv[i], "-d") == 0 && i < argc-1)
        {
            float dissolveFrac = (float)atof(argv[++i]);
            if (dissolveFrac < 0.0f || dissolveFrac > 1.0f)
                cerr << "WARNING: Ignoring dissolveFrac (out of range [0..1])\n";
            else
                info.dissolveFrac = dissolveFrac;
        }
        else if (strcmp(argv[i], "-x") == 0 && i < argc-1)
        {
            int temp = atoi(argv[++i]);
            if (temp <= 0)
                cerr << "WARNING: Ignoring output width (must be >0)\n";
            else
                info.outWidth = temp;
        }
        else if (strcmp(argv[i], "-y") == 0 && i < argc-1)
        {
            int temp = atoi(argv[++i]);
            if (temp <= 0)
                cerr << "WARNING: Ignoring output height (must be >0)\n";
            else
                info.outHeight = temp;
        }
        else if (strcmp(argv[i], "-a") == 0 && i < argc-1)
            info.morphParamA = (float)atof(argv[++i]);
        else if (strcmp(argv[i], "-b") == 0 && i < argc-1)
            info.morphParamB = (float)atof(argv[++i]);
        else if (strcmp(argv[i], "-p") == 0 && i < argc-1)
            info.morphParamP = (float)atof(argv[++i]);
    }

    return 0;
}



// MAIN -- the big one!
int main(int argc, char *argv[])
{    
    // process the command line arguments
    MorphInfo info;
    if (process_args(argc, argv, info) != 0)
    {
        exit(-1);
    }

    // other initialization
    build_sqrt_table();

    // morph!
    CImageMorpher morpher;
    morpher.SetNumThreads(4);

    if (!movieMode)
    {
        int err = morpher.Morph(info);
        if (err != 0)
        {
            cerr << "Morphing error!\n";
            exit(-1);
        }

        // Save the output image
        if (imgOut.Save(filenameOut) != 0)
        {
            cerr << "ERROR: error saving output file to " << filenameOut << endl;
            exit(-1);
        }
    }
    else
    {
        int encHandle;
        Revel_Error revError = Revel_CreateEncoder(&encHandle);
        
        Revel_Params params;
        params.width = info.outWidth;
        params.height = info.outHeight;
        params.frameRate = frameRate;
        params.quality = quality;
        params.codec = REVEL_CD_XVID;
        revError = Revel_EncodeStart(encHandle, filenameOut.c_str(), &params);
        if (revError != REVEL_ERR_NONE)
        {
            cerr << "ERROR: Could not open movie file " << filenameOut << endl;
            exit(-1);
        }

        Revel_VideoFrame frame;
        frame.width = params.width;
        frame.height = params.height;
        frame.pixelFormat = REVEL_PF_RGBA;
        frame.bytesPerPixel = 4;
        for(int i=0; i<numFrames; ++i)
        {
            float frac = (float)(i) / (float)(numFrames-1);
            info.warpFrac = frac;
            info.dissolveFrac = frac;
            morpher.Morph(info);
            frame.pixels = imgOut.GetPixels();
            int frameBytes;
            revError = Revel_EncodeFrame(encHandle, &frame, &frameBytes);
            cerr << "Frame " << i+1 << " of " << numFrames << ": "
                << frameBytes << " bytes\n";
        }
        revError = Revel_EncodeEnd(encHandle);
        Revel_DestroyEncoder(encHandle);
    }
    return 0;
}
